
=head1 XINABSE

=head2 common.pl

This include contains functions used by the spider

=cut

################################################################################

use LWP::UserAgent;
use HTML::TokeParser;
use HTML::Entities;
use HTML::LinkExtor;
use URI::URL;
use Getopt::Long;
use DBI;

################################################################################

# The following code checks if this script runs in Perl newer than 5.8 with
# internal Unicode strings by checking for the availability of the Encode
# module.
# Thanks to Benjamin Goldberg <goldbb2@earthlink.net> in comp.lang.perl.misc
BEGIN {
  my $have_encode = eval {
     require Encode;
     import  Encode qw(encode);
     return 1;
  } || 0;
  *HAVE_ENCODE = sub () { $have_encode };
}

################################################################################

=head3 getDBH

returns a new databasehandle

deprecated - use doSQL instead.

=cut

sub getDBH(){
  #takes global config options
  DBI->connect("dbi:mysql:".$CFG{db_database}.":".$CFG{db_server},
                $CFG{db_user},$CFG{db_password}) || die("Can't connect to DB");
}

=head3 doSQL I<query>

Takes a a SQL statement as argument and executes it. If no global database
handle is available it is created. Returns twodimensional array with the
result set. in [0][0] is the insert ID on a insert statement.

=cut

sub doSQL($){
  my $query = $_[0];
  my $result;
  #$_dbh is global!
  if(!defined($_dbh) || !$_dbh->ping()){
    $_dbh = DBI->connect("dbi:mysql:".$CFG{db_database}.":".$CFG{db_server},
                        $CFG{db_user},
                        $CFG{db_password}) || die("Can't connect to DB");
  }

  if($query =~ m/^SELECT/i){
    $result = $_dbh->selectall_arrayref($query);
    if($_dbh->err){
      debug(128,$_dbh->errstr);
      debug(128,$query);
      return undef;
    }else{
      return $result;
    }
  }else{
    unless($_dbh->do($query)){
      debug(128,$_dbh->errstr);
      debug(128,$query);
      return undef;
    }else{
      $result->[0][0] = $_dbh->{'mysql_insertid'}; # 0 for non inserts
      return $result;
    }
  }
}

=head3 utf8_quote I<$string>

quotes a string for database and converts perl 5.8+ Unicodestrings
back to latin1

returns returns quoted string

=cut

sub utf8_quote($){
  my $astring = shift;
  return 'NULL' unless(defined($astring));
  #perls behaviour on strings changed (utf8) this converts back to latin 1
  $astring = encode("iso-8859-1", $astring) if HAVE_ENCODE;
  $astring =~ s/(['"\\\0])/\\$1/g;
  return "'".$astring."'";
}

=head3 storeKeyword I<$keyword,$weight,$siteid>

Inserts a keyword for a site into the database

returns keyid

=cut

sub storeKeyword($$$){
  my ($keyword,$weight,$siteid) = @_;
  my ($SELECT,$result,$keyid);
  $keyword = utf8_quote($keyword);
  $siteid  = utf8_quote($siteid);

  my $INSERT = "INSERT IGNORE INTO xinabse_keywords
                   SET twoletters = LOWER(SUBSTRING($keyword,1,2)),
                       keyword = LOWER($keyword)";
  doSQL($INSERT);                       
  my $SELECT = "SELECT keyid
                 FROM xinabse_keywords
                WHERE twoletters = LOWER(SUBSTRING($keyword,1,2))
                  AND keyword = LOWER($keyword)";
  $result = doSQL($SELECT);
  $keyid  = $result->[0][0];
  $keyid  = utf8_quote($keyid);

  #check if keywordweight for this site exists
  $SELECT = "SELECT weight
               FROM xinabse_hits
              WHERE siteid = $siteid
                AND keyid = $keyid";
  $result = doSQL($SELECT);
                
  if(defined($result->[0][0])){
    doSQL("UPDATE xinabse_hits
              SET weight = weight+$weight
            WHERE siteid = $siteid
              AND keyid = $keyid");
  }else{
    doSQL("INSERT INTO xinabse_hits
              SET weight = $weight,
                  siteid = $siteid,
                   keyid = $keyid");
  }
  return $keyid;
}

=head3 storeSite I<$url,$lastmod,$title,$level[,$extract]>

Inserts (updates) a site and its data into the database

returns siteid

=cut

sub storeSite($$$$$){
  my ($url,$lastmod,$title,$level,$extract) = @_;
  my ($domain,$path,$page) = splitURL($url);
  $title   = utf8_quote($title);
  $level   = utf8_quote($level);
  $extract = utf8_quote($extract);
  $domain  = utf8_quote($domain);
  $path    = utf8_quote($path);
  $page    = utf8_quote($page);

  my $SET = "SET lastmodified = FROM_UNIXTIME($lastmod),
                 title = $title,
                 extract = $extract,
                 domain = $domain,
                 path = $path,
                 page = $page,
                 level = $level";

  my $SELECT = "SELECT siteid
                  FROM xinabse_sites
                 WHERE domain = $domain
                   AND path = $path
                   AND page = $page";
  my $result = doSQL($SELECT);

  if(defined($result->[0][0])){
    $siteid = $result->[0][0];
    doSQL("UPDATE xinabse_sites
                  $SET
            WHERE siteid = $siteid");
  }else{
    $result = doSQL("INSERT INTO xinabse_sites $SET");
    $siteid = $result->[0][0];
  }
  return $siteid;
}

=head3 deleteSite I<$url>

Deletes a site from the database

returns deleted siteid or 0 if site wasn't found

=cut

sub deleteSite($){
  my $url = $_[0];
  my ($domain,$path,$page) = splitURL($url);
  $domain  = utf8_quote($domain);
  $path    = utf8_quote($path);
  $page    = utf8_quote($page);

  my ($SELECT,$result,$siteid);
  $SELECT = "SELECT siteid
               FROM xinabse_sites
              WHERE domain = $domain
                AND path = $path
                AND page = $page";
  $result = doSQL($SELECT);
  if(defined($result->[0][0])){
    $siteid = $result->[0][0];
  }else{
    return 0;
  }
  doSQL("DELETE FROM xinabse_sites WHERE siteid = $siteid");
  doSQL("DELETE FROM xinabse_hits WHERE siteid = $siteid");
  debug(1,"$url DELETED");
  return $siteid;
}

=head3 splitURL I<$url>

splits URL into parts

returns domain,path,page

=cut

sub splitURL($){
  my $url = $_[0];
  $url = "http://".$url unless ($url =~ m/^http:\/\//);
  my $uri = URI->new($url);
  my $domain = $uri->host();
  my $temppath = $uri->path();
     $temppath = join("?",$uri->path(),$uri->query()) if($uri->query());
  my ($path,$page) = ($temppath =~ m/(.*\/)(.*?)$/);
  $path = '/' unless($path);
  $page = '' unless(defined($page));
  return ($domain,$path,$page);
}

=head3 fetch I<$url,$level[,startid,starturl]>

fetches URL and reachs the content to the parsers

returns nothing

=cut

sub fetch($$){
  my($url,$level,$startid,$starturl) = @_;
  debug(2,"fetching $url with level $level ...");
  my $ua = LWP::UserAgent->new;
  $ua->agent("XINABSE Spider/0.1");

  my($req,$res);
  
  #fetch header only
  $req = HTTP::Request->new(HEAD => $url);
  $res = $ua->request($req);
  if (!$res->is_success){
    debug (2,"document download failed");
    deleteSite($url);
    return;
  }elsif($res->headers->content_type !~ m/^text\//i){
    debug (2,"wrong contenttype");
    deleteSite($url);
    return;
  }

  #fetch document  
  $req = HTTP::Request->new(GET => $url);
  $res = $ua->request($req);
  if ($res->is_success){
    debug (2,"document downloaded ok");
    &parseDocument($res->headers,
                   $res->content,
                   $level,
                   $url,
                   $res->base,
                   $startid,
                   $starturl)
  }else{
    debug (2,"document download failed");
    deleteSite($url);
  }
}

=head3 countKeywords I<$hashref,$text [,$multiplier]>

counts all keywords (counts are multiplied with $multiplier) and stores the
number of found keywords in the given hash

returns nothing

=cut

sub countKeywords($$){
  my ($hashref,$text,$multi) = @_;
  $multi = 1 unless(defined($multi));
  $text = &stripSpecials($text);

  $CFG{minwordlength}=2 if ($CFG{minwordlength} < 2);
  $CFG{maxwordlength}=32 if ($CFG{maxwordlength} > 32);

  foreach my $word (split(" ",lc($text))){
    if( (length($word) >= $CFG{minwordlength}) &&
        (length($word) <= $CFG{maxwordlength}) &&
        ($word !~ m/^[0-9]+$/) && 
        (!isin($word,@STOPWORDS)) ){
      $$hashref{$word} = $$hashref{$word} + $multi;
    }
  }
}

=head3 parseDocument I<$header,$document,$level,$url,$base,$startid,$starturl>

parses header (HTTP::Headers) and document

returns nothing

=cut

sub parseDocument($$$$$$$){
  my ($header,$document,$level,$url,$base,$startid,$starturl) = @_;
  my $title = $header->title();
     unless($title){
       $document =~ m/<title>(.*?)<\/title>/sig;
       $title = $1;
       $title = $url unless ($title);  
     }
  my $lastmod = $header->last_modified();
  $lastmod = time() unless($lastmod);
  my($extract,%keywords,$keyword,$count);

  debug (8,"stripping html...");
  my $text = &stripHTML($document);

  if($CFG{saveextract}){
    $extract = substr($header->header('X-Meta-Description'),0,$CFG{extractsize});
    $extract = substr($text,0,$CFG{extractsize}) unless($extract);
  }

  debug (8,"storing sitinfo...");
  my $siteid = &storeSite($url,$lastmod,$title,$level,$extract);
  
  deleteSiteHits($siteid);
  
  debug (8,"finding keywords...");
  &countKeywords(\%keywords,$title,$CFG{titleweight});
  &countKeywords(\%keywords,$header->header('X-Meta-Keywords'),$CFG{metakeywordsweight});
  &countKeywords(\%keywords,$header->header('X-Meta-Description'),$CFG{metadescriptionweight}); 
  &countKeywords(\%keywords,$text);

  while(($keyword,$count)=each(%keywords)){
    &storeKeyword($keyword,$count,$siteid);
  }
  
  if($level){
    debug (8,"finding links...");
    $startid  = $siteid unless($startid);
    my @parts = splitURL($url);
    unless ($starturl){
      $starturl = "http://".$parts[0].@parts[1];
    }
    findLinks($document,$base,$level-1,$startid,$starturl,$parts[0]);
  }
}

=head3 stripHTML I<$html>

strips tags and newlines from HTML

returns textonly

=cut

sub stripHTML($){
  my $html = $_[0];
  my $text;
  debug (16,"document is ".length($html)." bytes...");

  #strip comments and noindextags
  $html =~ s/<noindex>.*?<\/noindex>//sig;
  $html =~ s/<!--UdmComment-->.*?<!--\/UdmComment-->//sig;
  $html =~ s/<!--noindex_start-->.*?<!--noindex_end-->//sig;
  $html =~ s/<!--htdig_noindex-->.*?<!--\/htdig_noindex-->//sig;
  $html =~ s/<!--stopindex-->.*?<!--startindex-->//sig;
  $html =~ s/<!--.*?-->//sig;
  $html =~ s/<STYLE.*?>.*?<\/STYLE>//sig;
  $html =~ s/<SCRIPT.*?>.*?<\/SCRIPT>//sig;
  $html =~ s/<HEAD.*?>.*?<\/HEAD>//sig;
  $html =~ s/<!DOCTYPE.*?>//sig;

  my $parser = HTML::TokeParser::->new( \$html);
  while (my $token = $parser->get_token) {
    next unless $token->[0] eq 'T';
    $text .= $token->[1];
  }

  $text = decode_entities($text);
  $text = encode_entities($text,"\200-\377");
  $text = decode_entities($text);
  $text =~ s/\s+/ /g;

  debug (16,"document stripped to ".length($text)." bytes.");
  return $text;
}

=head3 stripSpecials I<$text>

strips special chars and lowercase's

returns textonly

=cut

sub stripSpecials($){
  my $text = $_[0];

  # decode html entities
  $text = decode_entities($text);
  $text = encode_entities($text,"\200-\377");
  $text = decode_entities($text);
  #remove whitespaces
  $text =~ s/\s/ /gs;
  # remove any special char (nonletter latin1 chars)
  $text =~ s/[^0-9A-Za-z\xC0-\xFF]/ /gs;
  #remove doublespaces
  $text =~ s/  +/ /g;
  #lowercase
  $text = lc($text);

  return $text;
}


=head3 stripHTML I<$siteid>

removes all hits for a site

returns nothing

=cut

sub deleteSiteHits($){
  my $siteid = $_[0];
  doSQL("DELETE FROM xinabse_hits
          WHERE siteid = $siteid");
}

=head3 findLinks I<$html,$base,$level,$startid,$starturl,$domain>

removes finds all links and puts them into db

returns nothing

=cut

sub findLinks($$$$$$){
  my ($document,$base,$level,$startid,$starturl,$domain) = @_;
  my @links;
  #use LinkExtor to get <A HREF...>s
  my $le = HTML::LinkExtor->new(sub{
                                    my($tag, %attr) = @_;
                                    return unless ($tag eq 'a' || $tag eq 'frame');
                                    push(@links, values %attr);
                                   }
                               );
  $le->parse($document);
  # Expand all URLs to absolute ones
  @links = map { $_ = url($_, $base)->abs; } @links;
 
  #place links in database 
  LINK: foreach my $url (@links){
          next LINK unless ($url =~ m/^\Q$starturl/i); #stay below starturl
          foreach my $check (@{${$CFG{exclude}}{'*'}}){
            if($url =~ m/$check/i){
              debug (32,"Skipped $url because '$check' matched");
              next LINK;
            }
          }
          foreach my $check (@{${$CFG{exclude}}{lc($domain)}}){
            if($url =~ m/$check/i){
              debug (32,"Skipped $url because '$check' matched");
              next LINK;
            }
          }
          addURL($url,$level,$startid);
        } 
}

=head3 addURL I<$url,$level[,$startid]>

adds a new url.

returns nothing

=cut

sub addURL($$){
  my ($url,$level,$startid) = @_;
  my ($domain,$path,$page) = splitURL($url);
  $domain  = utf8_quote($domain);
  $path    = utf8_quote($path);
  $page    = utf8_quote($page);
  $startid = utf8_quote($startid);

  #check if site exists
  $SELECT = "SELECT siteid,level
               FROM xinabse_sites
              WHERE domain = $domain
                AND path = $path
                AND page = $page";
  $result = doSQL($SELECT);
  #if not put in DB
  if(!defined($result->[0][0])){
    doSQL("INSERT INTO xinabse_sites
              SET lastmodified = FROM_UNIXTIME(0),
                  lastcheck = FROM_UNIXTIME(0),
                  title = 'not spidered yet',
                  domain = $domain,
                  path = $path,
                  page = $page,
                  level = $level,
                  startid = $startid");
    &debug(1,"$url added with level $level and startid $startid");
  }else{
    #else update level if we have a bigger one it is starturl
    if (($level > $result->[0][1]) || $startid eq "NULL"){
      my $siteid = $result->[0][0];
      doSQL("UPDATE xinabse_sites
                SET level = $level,
                    lastcheck = FROM_UNIXTIME(0)
              WHERE siteid = $siteid");
      &debug(1,"$url changed level to $level and startid $startid");
    }else{
      &debug(1,"$url exists");
    }
  }
}

=head3 debug I<$verbositylevel,$message>

prints a message if $VERBOSITY is right

returns nothing

=cut

sub debug ($$){
  my ($lvl,$msg) = @_;
  
  if ($CFG{verbose} & $lvl){
    print STDERR "$msg\n";
  }
}

=head3 clearDB

deletes everything in database

returns nothing

=cut

sub clearDB(){
  doSQL("DELETE FROM xinabse_hits");
  doSQL("DELETE FROM xinabse_keywords");
  doSQL("DELETE FROM xinabse_sites");
}


=head3 getOutdatedURL I<$reindexperiod>

fetches an url that is older than $reindexperiod (in hours)

returns true|false

=cut

sub getOutdatedURL($$){
  my ($hours,$aryref) = @_;
  my $seconds = 3600*$hours;
  my $SELECT = "SELECT A.siteid,A.level,A.startid,
                       CONCAT('http://',A.domain,A.path,A.page),
                       CONCAT('http://',B.domain,B.path)
                  FROM xinabse_sites A LEFT JOIN xinabse_sites B
                    ON A.startid = B.siteid
                 WHERE UNIX_TIMESTAMP(A.lastcheck) < UNIX_TIMESTAMP() - $seconds
              ORDER BY A.lastcheck, A.level DESC
                 LIMIT 1";
  my $result = doSQL($SELECT);
  if(!defined($result->[0][0])){
    $$aryref[0]=undef;
    $$aryref[1]=undef;
    $$aryref[2]=undef;
    $$aryref[3]=undef;
    return 0;
  }else{
    my $siteid = $result->[0][0];
    my $level  = $result->[0][1];
    my $startid = $result->[0][2];
    my $url    = $result->[0][3];
    my $starturl= $result->[0][4];
        
    $$aryref[0]=$url;
    $$aryref[1]=$level;
    $$aryref[2]=$startid;
    $$aryref[3]=$starturl;
    return 1;
  }
}

=head3 markURL I<$url>

updates the lastcheck field of an URL to make sure the next run
doesn't pick that url again:

returns nothing

=cut

sub markURL($){
  my $url = $_[0];
  my($domain,$path,$page) = splitURL($url);
  
  $domain = &utf8_quote($domain);
  $path = &utf8_quote($path);
  $page = &utf8_quote($page);
  
  doSQL("UPDATE xinabse_sites
            SET lastcheck = NOW()
          WHERE domain = $domain
            AND path = $path
            AND page = $page");
}


=head3 loadStopwords I<$file>

loads stopwords from $file

returns @stopwords

=cut

sub loadStopwords($){
  my $file = $_[0];
  
  open(FILE,$file) || debug(4,"couldn't load stopwords from $file");
  my @temp = <FILE>;
  close(FILE);
  
  my @stopwords;
  foreach my $line (@temp){
    $line =~ s/#.*$//;
    $line =~ s/^\s+//;
    $line =~ s/\s+$//;
    next if ($line eq '');
    push(@stopwords,$line);
  }
  return @stopwords;
}

=head3 isin I<$needle,@haystack>

checks if $needle is in @haystack

returns true|false

=cut

sub isin($@){
  my $needle = shift;
  chomp($needle);
  my @haystack = @_;
  foreach my $row (@haystack){
    return 1 if ($needle eq $row);
  }
  return 0;
}

=head3 printStatistics

prints statistical data

returns nothing

=cut

sub printStatistics(){
  my($SELECT,$result,$keyc,$sitek,$swc);
  
  
  $SELECT = "SELECT count(*) FROM xinabse_keywords";
  $result = doSQL($SELECT);
  $keyc =  $result->[0][0];
  
  $SELECT = "SELECT count(*) FROM xinabse_sites";
  $result = doSQL($SELECT);
  $sitec =  $result->[0][0];

  $swc = scalar(@STOPWORDS);
  
  print "-"x80;
  print "\n$swc stopwords $keyc keywords $sitec pages\n";
  print "-"x80;
  print "\n";
  exit;
}

=head3 showMostCommon I<$limit>

show most common $limit keywords

returns nothing

=cut

sub showMostCommon($$){
  my $limit = $_[0];
  my $SELECT = "SELECT count(B.keyid),B.keyid, A.keyword
                  FROM xinabse_keywords A, xinabse_hits B
                 WHERE A.keyid = B.keyid
                 GROUP BY keyid
              ORDER BY 1 DESC
              LIMIT $limit";
  my $result = doSQL($SELECT);
  print "count\tkeyid\tkeyword\n";
  my $row=0;  
  while (defined($result->[$row][0])){
    print $result->[$row][0];
    print "\t";
    print $result->[$row][1];
    print "\t";
    print $result->[$row][2];
    print "\n";
    $row++;
  }
  exit;
}

=head3 showStartURLs

prints all startURLs

returns nothing

=cut

sub showStartURLs(){
  my $SELECT = "SELECT CONCAT('http://',domain,path,page)
                  FROM xinabse_sites
                 WHERE startid IS NULL";
  my $result = doSQL($SELECT);
  my $row=0;  
  while (defined($result->[$row][0])){
    print $result->[$row][0];
    print "\n";
    $row++;
  }
  exit;
}


=head3 readConfig I<$config>

reads a config file

returns hash with config data

=cut

sub readConfig($){
  my $cfgfile = $_[0];
  open (FILE,$cfgfile) || die("could not open $config");
  my @file = <FILE>;
  close (FILE);

  my ($name,$value);
  my %config;
  $config{exclude}={}; #prepare anonymous hashref
  my $section='';

  foreach my $line (@file){
    $line=~ s/#.*$//;
    $line=~ s/^\s*//;
    $line=~ s/\s*$//;
    next if ($line eq '');
    if($line =~ m/<EXCLUDE\s+(.*?)\s*>/i){

      $section = $1;
      ${$config{exclude}}{lc($section)} = []; #prepare anonymous arrayref
      next;
    }
    if($line =~ m/<\/EXCLUDE>/i){
      $section = '';
      next;
    }
    if($section){
      push(@{${$config{exclude}}{lc($section)}},$line);
    }else{
      ($name,$val) = split("=",$line);
      $val=~ s/^\s*//;
      $name=~ s/\s*$//;
      $config{lc($name)}=$val;
    }
  }

  return %config;
}


=head3 cleanUp

removes duplicates and senseless data from the database

returns nothing

=cut

sub cleanUp(){
  my($SELECT,$result,$row,$siteid,$keyid);
  
  # delete duplicate sites (this may occure because MySQLs lack of transaction
  # support in standard tableformats) and the parallel nature of the spider
  $SELECT = "SELECT A.siteid
               FROM xinabse_sites A, xinabse_sites B
              WHERE A.domain = B.domain
                AND A.path = B.path
                AND A.page = B.page
                AND A.siteid != B.siteid
           ORDER BY A.siteid DESC
              LIMIT 1";
  $result = doSQL($SELECT);
  while(defined($result->[0][0])){
    $siteid = $result->[0][0];
    doSQL("DELETE FROM xinabse_sites WHERE siteid = $siteid");
    debug(64,"removed duplicate site with id $siteid");
    $result = doSQL($SELECT);
  }
  
  #delete hits for non existant sites
  $SELECT = "SELECT A.siteid,A.keyid
               FROM xinabse_hits A LEFT JOIN xinabse_sites B
                 ON A.siteid = B.siteid
              WHERE B.siteid IS NULL";
  $result = doSQL($SELECT);
  $row=0;
  while(defined($result->[$row][0])){
    $siteid = $result->[$row][0];
    $keyid  = $result->[$row][1];
    doSQL("DELETE FROM xinabse_hits
                 WHERE siteid = $siteid
                   AND keyid = $keyid");
    debug(64,"removed hit for nonexistant site $siteid");
    $row++
  }
  
  #delete keywords without any hits
  $SELECT = "SELECT A.keyid
               FROM xinabse_keywords A LEFT JOIN xinabse_hits B
                 ON A.keyid = B.keyid
              WHERE B.keyid IS NULL";
  $result = doSQL($SELECT);
  $row=0;
  while(defined($result->[$row][0])){
    $keyid  = $result->[$row][0];
    doSQL("DELETE FROM xinabse_keywords
                 WHERE keyid = $keyid");
    debug(64,"removed nohit keyword $keyid");
    $row++;
  }
}


=head3 deleteSite I<url>

Deletes a startURL and all dependant sites and exits.

returns nothing

=cut

sub deleteSite($){
  my $url = $_[0];
  my ($domain,$path,$page) = splitURL($url);
  $domain = utf8_quote($domain);
  $path   = utf8_quote($path);
  $page   = utf8_quote($page);
  my $SELECT = "SELECT siteid
                  FROM xinabse_sites
                 WHERE domain = $domain
                   AND path = $path
                   AND page = $page";
  my $result = doSQL($SELECT);
  if(defined($result->[0][0])){
    my $siteid = $result->[0][0];
    doSQL("DELETE FROM xinabse_sites
            WHERE siteid = $siteid");
    doSQL("DELETE FROM xinabse_sites
            WHERE startid = $siteid");
    cleanUp();
  }
  exit;
}


### return true
return 1;
