#!/usr/bin/perl

require "common.pl";
%CFG = &readConfig("xinabse.conf");

GetOptions('add=s'=>\$CFG{starturl},
           'level=i'=>\$CFG{startlevel},
           'delete-all'=>\$CFG{deleteall},
           'delete=s'=>\$CFG{delete},
           'help|?'=>\$CFG{help},
           'statistics'=>\$CFG{printstats},
           'reindex-after=i'=>\$CFG{reindexperiod},
           'show-most-common=i'=>\$CFG{showkeys},
           'bookmarks=s'=>\$CFG{bookmarks},
           'print'=>\$CFG{showstarts},
           'maxprocesses=i'=>\$CFG{maxprocesses},
           'verbosity=i'=>\$CFG{verbose},
           'cleanup'=>\$CFG{cleanup}) || exit 1;


&help if($CFG{help});                              # help and exit
&deleteSite($CFG{delete}) if($CFG{delete});        # delete site and exit
&deleteAll() if($CFG{deleteall});                  # delete all and exit
showMostCommon($CFG{showkeys}) if($CFG{showkeys}); # show most common words and exit
@STOPWORDS = &loadStopwords($CFG{stopwordfile});   # prepare stopwords
printStatistics() if($CFG{printstats});            # print Stats and exit
showStartURLs() if ($CFG{showstarts});             # show startURLs and exit


&addURL($CFG{starturl},$CFG{startlevel}) if ($CFG{starturl}); # add starturl
&readBookmarkfile($CFG{bookmarks}) if ($CFG{bookmarks});      # add bookmarks

if($CFG{cleanup}){
  cleanUp(); #run cleanup
  exit;
}

&run(); #spider



################################################################################
# subs

sub deleteAll(){
  my $safety;
  print "This will delete ALL sites from the Database!\n";
  print "Type 'yes' if you really want to do this: ";
  $safety = <STDIN>;
  chomp($safety);
  if($safety =~ m/^yes$/){
    clearDB();
    print STDERR "Database cleared\n";
  }
  exit;
}

sub help(){
  #print help and exit
  print <<end;
  SYNTAX:
    xinabse-spider.pl <options>
  
  OPTIONS:
  
  --add <url>                 add new URL to spider
  --level <level>             set level for above URL to <level>
  --delete-all                deletes all keywords and sites
  --delete <url>              delete a startURL and all depending sites
  --help | -?                 this helpscreen
  --statistics                print statistics and exit
  --reindex-after <hours>     Set reindexperiod to <hours>
  --show-most-common <count>  print <count> most common keywords and exit
  --bookmarks <bookmarkfile>  read starturls (level 0) from netscape bookmarks
  --print                     print all startURLs in the database
  --maxprocesses <count>      number of concurent processes
  --verbosity <level>         set verbosity (see config)
  --cleanup                   remove duplicates and exit
  
  See documentation for more info
end
  exit;
}

sub run(){
  my @site;
  my $running = 0;
  while(getOutdatedURL($CFG{reindexperiod},\@site) || $running){
    if ($running < $CFG{maxprocesses} && defined($site[0])){
      if(fork()){ #parent
        markURL(@site[0]); #mark URL as checked
        $running++;
      }else{ #child
        #fetch(url,level,startid,starturl);
        fetch($site[0],$site[1],$site[2],$site[3]);
        exit;
      }
    }else{
      wait();
      $running--;
    }
  }
  #wait for remaining processes (should never happen)
  while(wait != -1){}
  #cleanup
  cleanUp();
}

sub readBookmarkfile($){
  my $bookmarks = $_[0];
  open(FILE,$bookmarks) || die("Could not open $bookmarks");
  my @contents = <FILE>;
  close FILE;
  my $content = join("",@contents);
  &findLinks($content,'',0,undef,'http://');
}
